'****************************************************************
'*  Name    : MENU_INC.BAS                                      *
'*  Author  : John Barrat                                       *
'*  Notice  : Copyright (c) 2005 John Barrat                    *
'*          : All Rights Reserved                               *
'*  Date    : 18/04/2005                                        *
'*  Version : 1.0                                               *
'*  Notes   : Generalised User Interface using 4 buttons & LCD  *
'****************************************************************
' NOTE  -   This INC File provides a basic menu navigation and control mechanism
'           to implement the application specific User Interface functions

' These routines provide a basic user interface using 4 push buttons and the LCD.
' The button functions are:

'   Menu    This button forces the system into the menu mode.  In menu mode it switches
'           between main menu items.
'   Inc/Dec These 2 buttons allow navigation within a menu item and to adjust parameters
'           and make selections
'   Set     This button is used to enter or accept a selection.

' Depressions of the keys or menu timeout will make calls to the following external routines:

'   ProcMainMenu  / ProcCfgMenu     These proceedures handles the Top menu items/Default Display
'   ProcMainSet   / ProcCfgSet      These proceedures handle the Set actions
'   ProcMainIncDec/ ProcCfgIncDec   These proceedure handle the IncDec Actions
'   
' The above routines must be declared and written in the Main Program

' The system allows for up to 15 top level menu items which are invoked by successive presses
' of the Menu button.  Sub menu items can be selected in one of 2 ways:

'   a) Using a fixed predetermined sequence which is advanced using the Set button
'   b) Using the Inc/Dec button to move to the next or previous menu item 
 
' Method a) is used when there are values which have to be adjusted using the Inc/Dec buttons. 
'   e.g. Setting time and date values

' Method b) is used to display a number of items using the Inc/Dec to navigate between the items

' Menu sequencing uses the MenuAct Byte to determine the menu item selected
' The upper 4 bits determine the Main Menu while the lower 4 bits determine the Sub menu item

'   MenuAct value of $00 is menu inactive  - i.e. the default display
'   MenuAct value of $10 is main menu 1 active (no sub menus)
'   MenuAct value of $11 is main menu 1, submenu 1 active

' To return to the default (no menu) call CancelMenu which will reset MenuAct to $00 and will 
' return the display to the normal mode.  This could be invoked under a menu item or will occur 
' automatically after a timeout from the last key pressed.

' 2 possible menu sets are supported, selected by the state of the Config Bit
' The MenuHold Bit is used to prevent the menu button from allowing the user to break out of an 
' incomplete sequence.  E.g. Clock setup.
' Similarly the SetInhib button is used to inhibit the use of the SET button where it is 
' inappropriate to do so.  There are no facilities for disabling INC/DEC.

' Three main subroutines are provided to handle the actions taken under the buttons and to 
' control the navigation through the menus.  The subroutines are: 
 
'   MnuPressed      - Moves between Main Menu items and sets up the initial screen for each menu
'   SetPressed      - Actions selections made in a menu or moves to next sub menu item
'   IncDecPressed   - Moves between sub-menus or adjusts parameters

' Additional routines control the Menu timeout, report an error and acknowledge a command

' MENU_INC assumes a 2 line * 16 Char LCD display.  Generally the Top line would be used to
' Display the menu item and any user prompts, the bottom line would carry the parameters

' The MainMenu item titles must be added to the text messages at the end of main program
' These should be in the form:

'KeyScan 	Var PORTA	 ' Port used to scan matrix keypad  	   (Matrix Only)
'ScanMask	Con $03      ' Mask for scan lines - must start from 0 (Matrix Only)

'Device = 18F252

Dim Mnu_i	As Byte	          ' Internal variable used by MENU_INC

' Key Mapping
Dim Keys 	     As Byte      ' Each bit holds the state of individual button or Matrix value
Symbol MnuButton = Keys.MenuB ' Press for Menu Button
Symbol DecButton = Keys.DecB  ' Press for Decrement Button
Symbol IncButton = Keys.IncB  ' Press for Increment Button
Symbol SetButton = Keys.SetB  ' Press for Set/memorise Button

Symbol RepeatRate =	50	      ' Button Repeat rate

Dim ActPtr	    As Byte	      ' Pointer to menu branch - calcuated from MenuAct MS nibble -1
Dim MenuAct     As Byte       ' Menu/SubMenu pointer
Dim SubMenuCnt	As Byte	      ' Bottom nibble only - indicates number of sib menu items
Dim TimeOutCtr  As Byte       ' Current Timeout value
Dim TimeOutVal  As Byte       ' Timeout Value on Reset
Dim ErrCode     As Byte       ' Error code number echoed to LCD on error
Dim KeyRepCtr	As Byte       ' Key repeat rate counter 
Dim Mnu_TB1		As Byte       ' Temporary Byte

Dim MnuHold     As Bit        ' when true deactivates the Menu Button
Dim SetHold     As Bit        ' when true deactivates the Set button
Dim IncDecHold  As Bit        ' when true deactivates the IncDec Buttons
Dim ConfigMode  As Bit        ' When true defines configuration mode

Symbol RBIF   = INTCON.0   	  ' RB Port Interrupt Flag
Symbol RBIE   = INTCON.3   	  ' RB Port Change Interrupt Enable
Symbol TMR1ON = T1CON.0       ' Timer1 ON
Symbol TMR1CS = T1CON.1       ' Timer1 Clock Source Select
Symbol TMR1IF = PIR1.0 		  ' TMR1 Overflow Interrupt Flag bit
Symbol TMR1IE = PIE1.0		  ' TMR1 Interrupt enable

GoTo SkipMENU_Inc

Menu_Init:
    MenuAct = 0                      ' Reset the menu pointer
    TimeOutVal = 20                  ' Set the menu timeout value to 10Secs
    ConfigMode = False               ' Clear the configuration mode
    MnuHold = False
Return


#IFDEF IOC_HANDLER
KeyScan_Intr:
    Clear RBIF: Clear RBIE 	  		 ' Clear the IOC Interrupt flag and Turn Off Interrupt
    Keys = KeyIn & KeyMask           ' Save current value
	KeyRepCtr = 4  					 ' Set Debounce timeout to 80mS
GoTo INT_Code_Cont                   ' Return to top of Interrupt handler

KeyRepTMR:
    If Keys < KeyMask Then    		 ' If a key is pressed
       If (KeyIn & KeyMask) = Keys Then 
	   ' See if the same button is still being pressed
       	  Dec KeyRepCtr              ' Yes so decrement the repeat counter
       	  If KeyRepCtr = 0 Then      ' See if we have reached the repeat rate
   	      	 ADDImmd ProcessButtons	 ' Yes so process the key again
       	  	 KeyRepCtr = RepeatRate  ' Reset the KeyRepeat counter
	      EndIf
       EndIf
    Else
	    Set   RBIE                	 ' and turn the interrupt on change back on.
    EndIf
    'Do any other timer work here......
    
    Clear TMR1IF                     ' Clear the Timer 1 Overflow flag
GoTo INT_Code_Cont                   ' Return to top of Interrupt handler

#ELSE
' Keypad Scan - use this routine when using non-matrixed buttons (Max of 4 buttons)
ScanButtons:
    GoSub GetScan: If Keys = Mnu_i Then ExitScan
ScanRepeat:                                     ' Enter here to detect key repeats
	If Mnu_i > $01 Then                        	' See if a button is pressed
		Keys = Mnu_i                           	' Take a snapshot of the buttons in KeyScan
	  	DelayMS 10 : GoSub GetScan              ' wait for a while to debounce and recheck
	  	If Keys = Mnu_i Then ProcessButtons    	' If still pressed process action
	EndIf
    Keys = $FF									' Clear stored Keyvalue as it is now processed
ExitScan:
Return                                			' Back to Scheduler

GetScan:
    Mnu_i = KeyIn & KeyMask                    	' Retreive the the current key states
Return
#ENDIF


' Reports "OK" at current cursor position
ActionOK:
    Print mOK                              		' Tell user, "OK"
' Sounds an OK beep on the Speaker
BeepOK:
    Sound Speaker,[50,10,100,10]
    DelayMS 300                                 ' Leave time to see message on screen
Return

ActionFail:
    Pop ErrCode
    Print mFailed, " Err #", DEC2 ErrCode  		' Tell user, "Failed" with error code
' Sounds an Error beep on the Speaker
BeepErr:
    Sound Speaker,[25,20]
    DelayMS 300                                 ' Leave time to see message on screen
Return  

' This routine is called from the Main Program Control Loop or interupt service
' routine twice per second when the menu is active.
DecMenuTimer:
    Dec TimeOutCtr                 		 		' Decrement Menu TimeOut
    If TimeOutCtr > 0 Then Return               ' Return if not timed out  Test it for Timeout
CancelMenu:
    GoSub CancelMainMenu                    	' Call the Cancel activity for the current menu
	MenuAct = $0                            	' Clear down all menu activites and                                     	
 	REMTask DecMenuTimer, Rate1                 ' Remove this task
   	Cls: MnuHold = False                        ' Clear LCD and enable the menu button
    GoTo ProcMainMenu                           ' and return through ProcMainMenu

' Main User Input Processing Routines **********************************************************

ProcessButtons:
    TimeOutCtr = TimeOutVal                     ' reset the Timer to maximum timeout
    If Keys = ForceConfig Then               	' Check for Forced Config mode
       If MenuAct = 0 Then                      ' only allow it if not already in a menu
           ConfigMode = True                    ' set the config mode flag 
	   	   GoTo MnuPressed	   	                ' and display the first config menu item
	   Else
	   	   GoSub BeepErr			 			' Can't enter config from a menu.
		   Return
	   EndIf		
	Else
	   If MnuButton = Pressed Then              ' Check against the current state of the button
       	  GoTo MnuPressed                       ' Process the Menu Button
 	   Else
      	  If DecButton = Pressed Then IncDecPressed
          If IncButton = Pressed Then IncDecPressed  ' process the Inc and Dec Buttons
          If SetButton = Pressed Then SetPressed     ' process the Set Button
   	   EndIf
	EndIf
Return

SMALL_MICRO_MODEL = On
CalcPtr:                     ' Calculates the 0 based index into the menu control processes 
   ActPtr = (MenuAct >>4)  	 ' from the top 4 bits of the MenuAct byte
Return

' This routine calls the routines for the MENU button - note button repeat allowed
MnuPressed:
    If MnuHold = True Then 
       GoSub BeepErr                           ' Force user to complete Current action 
    Else
       If MenuAct = 0 Then ADDTask DecMenuTimer, Rate1' If Menu not active Start Timeout counter
        GoSub CancelMainMenu                	' Call the Cancel activity for the current menu
	   If ConfigMode = True Then GoSub CancelCfgMenu
	   Cls                        		        ' Clear LCD 
       MenuAct = (MenuAct + $10) & $F0     	' Increment MenuAct to next main menu value
	   If ConfigMode = False Then              ' Normal mode
       	  If MenuAct > (MnuMTNum <<4) Then MenuAct = $10 ' If last back to first menu                              
          GoSub CalcPtr                       ' Set pointer value for menu item
       	  GoSub ProcMainMenu              	' Call Normal menu process in Main program
 	   Else                                	' Config mode so only show System SetUp options
          If MenuAct > (MnuMCNum <<4) Then MenuAct = $20 ' If last back to first menu     
		' Note - Config mode resets to MenuAct = $20 to skip the initial Confirmation screen
  		  GoSub CalcPtr                       ' Set pointer value for menu item
      	  GoSub ProcCfgMenu                   ' Call Config menu process in Main program
       EndIf		
    EndIf
#IFNDEF IOC_HANDLER
    GoTo KeyRepeat                              ' Return to mainloop via Repeat delay
#ELSE
    Return
#ENDIF

' This routine calls the routines for the SET button - note Button repeat not allowed on SET
SetPressed:
    If SetHold = True Then                      ' Set disabled so beep error 
        GoSub BeepErr
    Else
        GoSub CalcPtr                           ' Set pointer value for menu item
        If ConfigMode = False Then              ' Normal Mode
            GoSub ProcMainSet                   ' Call Normal Set Action in Main Program
        Else                                    ' Config mode set
            GoSub ProcCfgSet                    ' Call Config Set Action in Main Program
        EndIf
    EndIf
Return' GoTo ExitScan                                   ' Back to the Scheduler

' This routine calls the routines for the INC/DEC button - note Button repeat allowed
IncDecPressed:
    If IncDecHold = True Then
        GoSub BeepErr                           ' IncDec Disabled so beep error
    Else
        GoSub CalcPtr                           ' Set pointer value for menu item
        If ConfigMode = False  Then             ' Normal Mode
            GoSub ProcMainIncDec                ' Call Normal IncDec Action in Main Program
        Else                                    ' Config Mode
            GoSub ProcCfgIncDec                 ' Call Config IncDec Action in Main Program
        EndIf
    EndIf
#IFNDEF IOC_HANDLER
KeyRepeat:
    DelayMS RepeatRate
GoTo ScanRepeat  								' See if key still pressed
#ELSE
Return
#ENDIF
SMALL_MICRO_MODEL = Off

' Cycles through the Sub Menu items resetting to 0 when it has reached the end.
' SubMenuCnt defines the number of menu items to be displayed in the cuurent menu.
' IncDec determines the direction based on the state of the Inc button
IncDecSubMenu:                                  ' Move to next submenu parameter
If IncButton = Pressed Then IncSubMenu
SMALL_MICRO_MODEL = On
DecSubMenu:                                     ' Decrement submenu
   If (MenuAct & $0F) =  0 Then MenuAct = MenuAct |(SubMenuCnt & $0F)
   Dec MenuAct
Return

IncSubMenu:                                     ' Increment Sub Menu
   Inc MenuAct
   If (MenuAct & $0F) = SubMenuCnt Then MenuAct = MenuAct & $F0
Return
SMALL_MICRO_MODEL = Off

SkipMENU_Inc:
